//====================================================================================//
//                                                                                    //
//                       Class Harwell-Boeing Format Matrix                           //
//                                                                                    //
//====================================================================================//
//  This File:   hbf.C                      Language: C++  (xlC and CC)               //
//  Software:    PEPS 2002                   Advisor: Brigitte.Plateau@imag.fr        //
//  Error Codes: 3600                                                                 //
//====================================================================================//
//  Creation:    19/dec/95                        by: Paulo.Fernandes@imag.fr         //
//  Last Change:  07/feb/04                        by: Ihab.Sbeity@imag.fr             //
//====================================================================================//
#include "peps.h"

     //--------------------------------------------------------------------------//
     //                          Static variable                                 //
     //--------------------------------------------------------------------------//
HBF HBF::hbf = HBF();

     //--------------------------------------------------------------------------//
     //                      It returns the 'i,j' element                        //
     //--------------------------------------------------------------------------//
rp HBF::coord(const int i, const int j) const
{
#ifdef _PEPS_DEBUG_
  if ((i >= dimension) || (j >= dimension))
    Programming_Error("invalid coordinates", 3601)
#endif
 rp answer = zero;

 for (int k=columns[j]; k<columns[j+1]; k++)   // it searches the element in its column
   if (rows[k] == i) {
     answer = elements[k];
     break;
 }
 return(answer);
}
     //--------------------------------------------------------------------------//
     //                 It returns the largest element (in modulo)               //
     //--------------------------------------------------------------------------//
rp HBF::biggest() const
{
 rp answer = zero;

 for (int i=0; i<nz_number; i++)
   if (answer < absv(elements[i]))
     answer = absv(elements[i]);
 return(answer);
}
     //--------------------------------------------------------------------------//
     //                 It multiplies all matrix elements by 'f'                 //
     //--------------------------------------------------------------------------//
void HBF::scale(const rp f)
{
 for (int i=0; i<nz_number; i++)
   elements[i] *= f;
}
     //--------------------------------------------------------------------------//
     //          It multiplies all matrix elements by the largest element        //
     //--------------------------------------------------------------------------//
void HBF::scale()
{
 rp factor = biggest();
#ifdef _PEPS_DEBUG_
  if (absv(factor) < close_to_zero)
    Programming_Error("scaling a zero matrix", 3602)
  else
#endif
 factor = one / factor;
 scale(factor);
}

     //--------------------------------------------------------------------------//
     //                            Transpose the matrix                          //
     //--------------------------------------------------------------------------//
void HBF::transpose()
{
 rp  * new_elem;
 int * new_rows;
 int * new_cols;
 int i, j, k, next;

 new_elem = new  rp[nz_number];
 new_rows = new int[nz_number];
 new_cols = new int[dimension+1];

 for (j=0; j<dimension+1; j++)                 // counts the elements by row
   new_cols[j] = 0;
 for (k=0; k<nz_number; k++)
   new_cols[rows[k]+1]++;
 for (j=1; j<dimension; j++)
   new_cols[j+1] += new_cols[j];

 for (j=0; j<dimension; j++)                   // copies the elements
   for (k=columns[j]; k<columns[j+1]; k++) {
     new_rows[new_cols[rows[k]]] = j;
     new_elem[new_cols[rows[k]]] = elements[k];
     new_cols[rows[k]]++;
   }
 for (j=dimension-1; j>=0; j--)
   new_cols[j+1] = new_cols[j];
 new_cols[0] = 0;

 delete[] elements;
 elements = new_elem;
 delete[] rows;
 rows     = new_rows;
 delete[] columns;
 columns  = new_cols;
}

     //--------------------------------------------------------------------------//
     //    Output on 'fout' file stream converting the row and columns indexes   //
     //--------------------------------------------------------------------------//
void HBF::convert(const peps_name file, const hbf_format format)
{
 ofstream fout;
 int      i;

 transpose();           // transpose the matrix - the result is always in c format
 if (format == C_format)
   Open_File(file, hbf_file, fout);
 else
   Open_File(file, dot_file, fout);
 fout << dimension << " " << nz_number << "\n";
 for (i=0; i<nz_number; i++)
   fout << elements[i] << "\n";
 if (format == C_format) {              // keep on 'c' format
   for (i=0; i<nz_number; i++)          //   (it have been converted on transpose)
     fout << rows[i]       << " ";
   fout << "\n";
   for (i=0; i<=dimension; i++)
     fout << columns[i]    << " ";
 }
 else {                                 // convert from 'c' to 'fortran' format
   for (i=0; i<nz_number; i++)          //   (adds one to each row and column
     fout << rows[i]+1     << " ";      //                                indexes)
   fout << "\n";
   for (i=0; i<=dimension; i++)
     fout << columns[i]+1  << " ";
 }
 fout << "\n";
 if (format == C_format)
   Close_File(file, hbf_file, fout);
 else
   Close_File(file, dot_file, fout);
 erase();
}
     //--------------------------------------------------------------------------//
     //                     Output on 'fout' file stream                         //
     //--------------------------------------------------------------------------//
void HBF::write() const
{
 ofstream fout;
 int      i;

 Open_File(file_name, hbf_file, fout);
 fout << dimension << " " << nz_number << "\n";
 for (i=0; i<nz_number; i++)
   fout << elements[i] << "\n";
 for (i=0; i<nz_number; i++)
   fout << rows[i]     << " ";
 fout << "\n";
 for (i=0; i<=dimension; i++)
   fout << columns[i]  << " ";
 fout << "\n";
 Close_File(file_name, hbf_file, fout);
}



     //--------------------------------------------------------------------------//
     //                     Input from 'fin' file stream                         //
     //--------------------------------------------------------------------------//
void HBF::read(const peps_name file)
{
 ifstream fin;
 int n, d, i;

 Open_File(file, hbf_file, fin);
 fin >> d;
 fin >> n;
 create(n, d);
 baptise(file);
 for (i=0; i<nz_number; i++)
   fin >> elements[i];
 for (i=0; i<nz_number; i++)
   fin >> rows[i];
 for (i=0; i<=dimension; i++)
   fin >> columns[i];
 Close_File(file_name, hbf_file, fin);
}
     //--------------------------------------------------------------------------//
     //              Memory space used by 'this' (in memory units)               //
     //--------------------------------------------------------------------------//
int HBF::mem_use() const
{
 int answer = 8;
 answer += nz_number * 8;
 answer += nz_number * 4;
 answer += (dimension+1) * 4;
 return(answer);
}

     //--------------------------------------------------------------------------//
     //                             Main Functions                               //
     //--------------------------------------------------------------------------//
     //                                                                          //
     //                     Generation from the Descriptor                       //
     //                                                                          //
     //    Input Parameters:                                                     //
     //                                                                          //
     //      - 'DSC::dsc'  the descriptor to be expanded;                        //
     //      - 'FTB::ftb'  the function table;                                   //
     //      - 'RSS::rss'  the reachable state space;                            //
     //      - 'name'      the file name of the new structure (the .hbf file);   //
     //                                                                          //
     //    Output Parameter:                                                     //
     //                                                                          //
     //      - 'this'      the sparse matrix corresponding to the descriptor;    //
     //                                                                          //
     //   This function generates a sparse Hardwell-Boeing format matrix (with   //
     // elements of the same column stored contiguously) corresponding to a      //
     // given SAN descriptor ('DSC::dsc').                                       //
     //                                                                          //
     //   IMPORTANT: for huge models this function may need HUGE memory space.   //
     // The memory allocation during of the elements generation is made by       //
     // individual calls of sparse vectors for each column. However, the final   //
     // storage of the matrix requires an extra temporary space to assembly the  //
     // sparse vectors of each column in a single vector. This function actually //
     // needs (at its final step) a memory space about twice the actual need to  //
     // store the matrix.                                                        //
     //--------------------------------------------------------------------------//
     //   Algorithm Principles:                                                  //
     //        - This function behaves like a complete aggregation of the main   //
     // descriptor. All automata of the model are aggregated into a single one   //
     // (all transitions become local to this single automata);                  //
     //        - The order of the resulting matrix will be reduced according to  //
     // the reachable state space size;                                          //
     //        - The numerical evaluation of each element of the resulting       //
     // matrixm is performed by the "evaluate a coordinate" member function of   //
     // the class descriptor ('rp dsc::coord(i,j);                               //
     //        - The descriptor may refers to a reduced state space (due to a    //
     // grouping process). In this, case its original information may be be      //
     // changed (different number and size of automata, remaining states, etc).  //
     //--------------------------------------------------------------------------//
     //   Algorithm: - the original state space is visited column by column;     //
     //                . if the column corresponds to a reachable state:         //
     //                  - for all elements in this column;                      //
     //                    . if its row corresponds to a reachable state:        //
     //                      - computes the transition rate;                     //
     //                      - store it in the temporary 'buf';                  //
     //                  - copies 'buf' vector in a sparse column structure;     //
     //              - copies all temporary sparse columns to 'this' structure;  //
     //--------------------------------------------------------------------------//
     //   Remark: this implementation performs three copies of each matrix       //
     // elements and requires a little bit less than twice the memory space of   //
     // the result. This waste of time and memory can be avoided if a good       //
     // estimation of the number of nonzero elements can be obtained.            //
     //--------------------------------------------------------------------------//
#ifdef _PEPS_DSC_
void HBF::generate(const peps_name file)
{
 erase();

 //----------------------------------------------------- Original descriptor information
 int    o_ss;      // the product state space (original state space size)
 int    o_i, o_j;  // coordinates of the rows and columns of the product state space
 aut_st o_sizes;   // the size of each original automaton
 aut_st o_state;   // the state vector (corresponding to the state in 'o_i')

 o_ss = RSS::rss.orig_size();           // the product state space size
 FTB::ftb.get_domain(o_sizes);          // the automata sizes
 o_state.create(FTB::ftb.aut());        // the state vector
 o_state.reset();                       //     (allocation and reset)

 //------------------------------------------------------ Current descriptor information
 aut_st d_sizes;   // the size of each automata in the descriptor
 aut_st d_i;       // the state vector (used to index the rows of the descriptor)
 aut_st d_j;       // the state vector (used to index the columns of the descriptor)

 DSC::dsc.get_s(d_sizes);               // gets the automata sizes
 d_j.create(DSC::dsc.aut());            // the state vector for the rows (allocation)
 d_i.create(DSC::dsc.aut());            // the state vector for the columns (allocation)

 //-------------------------------------------------------- Resulting matrix information
 int    r_i, r_j;  // coordinates for the reachable state space (and resulting matrix)
 int    k;         // counter for nonzero elements of the resulting matrix

 dimension = RSS::rss.reach_size();     // the reachable state space size

 //--------------------------------------------------------------- Temporary information
 rp  ** t_elem;   // temporary array for the sparse columns nonzero elements
 int ** t_rows;   // temporary array for the row positions of each nonzero element
 int  * t_nzeros; // temporary size (number of nonzero elements) of each sparse column
 rp   * buf;      // a buffer to store the visited colum (non sparse)


 t_elem = new rp*[dimension];                       // alocates the temporary arrays
 t_rows = new int*[dimension];                      //   for the rows and nonzeros
 t_nzeros = new int[dimension];                     //   elements, the size of each
 buf = new rp[dimension];                           //   column and the buffer column


 // Ihab ------------------------- Get colone number to print

 //------------------------------------------------------------ "Visit all columns" loop
 for (o_j=0, d_j.reset(), r_j=0; o_j<o_ss; o_j++) { // for each original state       - j
   if (RSS::rss.remains(o_j)) {                     //   if it is in the descriptor  - j
     if (RSS::rss.has(o_j)) {                       //     if it is reachable        - j
       t_nzeros[r_j] = 0;                                 // init. the number of nonzero
                                                          // elements of this column

       //--------------------------------------------------------- "Visit all rows" loop
       for (o_i=0, d_i.reset(), r_i=0; o_i<o_ss; o_i++) { // for each original state - i
         if (RSS::rss.remains(o_i)) {                     //   if it is in the desc. - i
           if (RSS::rss.has(o_i)) {                       //     if it is reachable  - i

             //--------------------------------------------- Computes the matrix element
             if (absv(buf[r_i] = DSC::dsc.coord(d_i, d_j, o_state)) > close_to_zero) {
                                                             // stores it in 'buf'
               t_nzeros[r_j]++;                              // counts nonzeros
	       //cout << "r_i: "<< r_i <<" r_j: "<< r_j <<" val: "<< buf[r_i] <<"\n";
	     }
             //-------------------------------------------------------------------------

             r_i++;                                       //     advance reach state - i
           }
           d_i.next(d_sizes);                             //   advance desc. state   - i
         }
         o_state.next(o_sizes);                           // advance original stace  - i
       }
       //-------------------------------------------------------------------------------

       t_elem[r_j] = new rp[t_nzeros[r_j]];             // allocates sparse column 'r_j'
       t_rows[r_j] = new int[t_nzeros[r_j]];
       for (r_i=0, k=0; r_i<dimension; r_i++)           // it transfers the computed
         if (absv(buf[r_i]) > close_to_zero) {          //   column from 'buf' to the
           t_elem[r_j][k] = buf[r_i];                   //   temporary sparse column
           t_rows[r_j][k] = r_i;                        //   (vectors 't_elem' and
           k++;                                         //    't_rows')

	 }
       r_j++;                                       //     advance reachable state   - j
     }
     d_j.next(d_sizes);                             //   advance descriptor state    - j
   }
   cout << "\r                                      "
        << "\rGenerating Column " << o_j << "/" << o_ss << " (pss) "
        << r_j << "/" << dimension << " (rss)";                             cout.flush();
 }                                                  // advance original state        - j
 //-------------------------------------------------------------------------------------
 delete[] buf;
 cout << "\r                                      "
      << "\rGenerating Column " << o_j << "/" << o_ss << " (pss) "
        << r_j << "/" << dimension << " (rss)\nFinishing ... ";       cout.flush();


 //----------------------------------------------------- Copies sparse columns to 'this'
 nz_number = 0;
 for (r_j=0; r_j<dimension; r_j++)            // counts the total number of nonzeros
   nz_number += t_nzeros[r_j];
 elements  = new rp[nz_number];               // allocates 'this' structures
 rows      = new int[nz_number];
 columns   = new int[dimension+1];
 nz_number = 0;
 for (r_j=0; r_j<dimension; r_j++) {          // for each column
   columns[r_j] = nz_number;                     // stores its first element index
   for (k=0; k<t_nzeros[r_j]; k++) {             // copies its elements
     elements[nz_number] = t_elem[r_j][k];
     rows[nz_number] = t_rows[r_j][k];
     nz_number++;
   }
   delete[] t_elem[r_j];                         // erases its sparse column vectors
   delete[] t_rows[r_j];
 }
 columns[r_j] = nz_number;                    // stores the end of the last column index

 delete[] t_elem;                             // erases temporary structures
 delete[] t_rows;
 delete[] t_nzeros;
 baptise(file); write();                      // stores 'this' name and save it!

}
#endif
