//====================================================================================//
//                                                                                    //
//                            Class Probability Vector                                //
//                                                                                    //
//====================================================================================//
//  This File:   vct_inl.h                  Language: C++  (xlC and CC)               //
//  Software:    PEPS 3.1                    Advisor: Brigitte.Plateau@imag.fr        //
//  Error Codes: 2300                                                                 //
//====================================================================================//
//  Creation:    19/dec/95                        by: Paulo.Fernandes@imag.fr         //
//  Last Change: 16/feb/00                        by: Paulo.Fernandes@imag.fr         //
//====================================================================================//

     //--------------------------------------------------------------------------//
     //                          Empty Constructor                               //
     //--------------------------------------------------------------------------//
inline VCT::VCT()
{
 dimension = 0;
 p_order.create();
 p_jumps.create();
 sizes.create();
}
     //--------------------------------------------------------------------------//
     //                          Sized Constructor                               //
     //--------------------------------------------------------------------------//
inline VCT::VCT(const int dim)
{
 dimension = dim;
 vec = new rp[dimension];
 p_order.create();
 p_jumps.create();
 sizes.create();
}
     //--------------------------------------------------------------------------//
     //                           Copy Constructor                               //
     //--------------------------------------------------------------------------//
inline VCT::VCT(const VCT & v)
{
 dimension = v.dimension;
 vec = new rp[dimension];
 for (int i=0; i<dimension; i++)
   vec[i] = v.vec[i];
 p_order.copy(v.p_order);
 p_jumps.copy(v.p_jumps);
 sizes.copy(v.sizes);
}
     //--------------------------------------------------------------------------//
     //                              Destructor                                  //
     //--------------------------------------------------------------------------//
inline VCT::~VCT()
{
 if (dimension != 0) {
   delete[] vec;
   dimension = 0;
 }
 p_order.erase();
 p_jumps.erase();
 sizes.erase();
 strcpy(file_name, "\0");
}

     //--------------------------------------------------------------------------//
     //                           Empty Intialization                            //
     //--------------------------------------------------------------------------//
inline void VCT::create()
{
 dimension = 0;
 p_order.create();
 p_jumps.create();
 sizes.create();
}
     //--------------------------------------------------------------------------//
     //                           Sized Intialization                            //
     //--------------------------------------------------------------------------//
inline void VCT::create(const int dim)
{
 if (dimension != dim) {
   erase();
   dimension = dim;
   vec = new rp[dimension];
 }
 p_order.erase();
 p_jumps.erase();
 sizes.erase();
}
     //--------------------------------------------------------------------------//
     //                       Copy and Sized Intialization                       //
     //--------------------------------------------------------------------------//
inline void VCT::copy(const VCT & v)
{
 create(v.dimension);
 for (int i=0; i<dimension; i++)
   vec[i] = v.vec[i];
 p_order.copy(v.p_order);
 p_jumps.copy(v.p_jumps);
 sizes.copy(v.sizes);
}
     //--------------------------------------------------------------------------//
     //                            Destruction                                   //
     //--------------------------------------------------------------------------//
inline void VCT::erase()
{
 if (dimension != 0) {
   delete[] vec;
   dimension = 0;
   strcpy(file_name, "\0");
 }
 p_order.erase();
 p_jumps.erase();
 sizes.erase();
}

     //--------------------------------------------------------------------------//
     //                          It assigns a file name                          //
     //--------------------------------------------------------------------------//
inline void VCT::baptise(const peps_name file)
{
 strcpy(file_name, file);
}
     //--------------------------------------------------------------------------//
     //                         It returns the file name                         //
     //--------------------------------------------------------------------------//
inline void VCT::name(peps_name file) const
{
 strcpy(file, file_name);
}
     //--------------------------------------------------------------------------//
     //                        The total number of states                        //
     //--------------------------------------------------------------------------//
inline int VCT::size() const
{
 return(dimension);
}
     //--------------------------------------------------------------------------//
     //                    The total number of states is zero                    //
     //--------------------------------------------------------------------------//
inline bool VCT::is_empty() const
{
 return(dimension == 0);
}

     //--------------------------------------------------------------------------//
     //                   The probability of the state 's'                       //
     //--------------------------------------------------------------------------//
inline rp VCT::get(const int s) const
{
#ifdef _PEPS_DEBUG_
  if (s >= dimension)
    Programming_Error("index out of range", 2301)
#endif
 return(vec[s]);
}
     //--------------------------------------------------------------------------//
     //             It puts the probability 'v' to the state 's'                 //
     //--------------------------------------------------------------------------//
inline void VCT::put(const int s, const rp v)
{
#ifdef _PEPS_DEBUG_
  if (s >= dimension)
    Programming_Error("index out of range", 2302)
#endif
 vec[s] = v;
}

     //--------------------------------------------------------------------------//
     //  It returns 'true' if the error between two vectors is smaller than 'e'  //
     //--------------------------------------------------------------------------//
     // Obs.: This function assumes that the absolute error is always less       //
     // restritive than the relative error. This assumption is true if 'this'    //
     // is a probability vector.                                                 //
     //--------------------------------------------------------------------------//
inline bool VCT::test(const VCT & v, const e_typ err, const rp e) const
{
#ifdef _PEPS_DEBUG_
  if (v.size() != dimension)
    Programming_Error("different sizes in compare", 2311)
#endif
 bool answer = true;
 rp   error  = zero;
 rp   diff;
 rp   max    = zero;

 for (int k=0; k<dimension; k++) {                // for all states
   diff = absv(vec[k] - v.vec[k]);                   // computes the absolute difference
   if (err == rel_ind_err)                           // if the precision is relative
     if (vec[k] > max)                                  // looks for the largest
       max = vec[k];                                    //   element
   if ((err == abs_ind_err) || (err == rel_ind_err)) // consider the error 
     error = diff;                                      // individually
   else                                                 //   or
     error += diff;                                     // accumulated
   if (error > e) {                                  // if the absolute error already
     answer = false;                                 //   is bigger than 'e'
     break;                                              // stop the test (ret. false)
   }
 }
 if ((err == rel_ind_err) && (answer))   // if the precision is relative and the
                                         //   absolute error never was bigger than 'e'
   answer = error/max < e;                  // consider the error relatively

 return(answer);
}
     //--------------------------------------------------------------------------//
     // It returns the sum of all elements, the largest element and its position //
     //--------------------------------------------------------------------------//
inline void VCT::info(rp & acc, rp & big, int & bpos, rp & sml, int & spos) const
{
 rp elem;

 big = zero;
 sml = one;
 acc = 0;
 for (int i=0; i<dimension; i++) {
   acc += vec[i];
   elem = absv(vec[i]);
   if (big < elem) {
     bpos = i;
     big = elem;
   }
   if ((elem > round_zero) && (sml > elem)) {
     spos = i;
     sml = elem;
   }
 }
}
